from collections.abc import Sequence
from typing import Literal as L, overload

import numpy as np
from numpy._typing import (
    NDArray,
    _ArrayLike,
    _ArrayLikeFloat_co,
    _ArrayLikeNumber_co,
    _Shape,
)
from numpy._typing._array_like import _DualArrayLike

__all__ = [
    "fft",
    "ifft",
    "rfft",
    "irfft",
    "hfft",
    "ihfft",
    "rfftn",
    "irfftn",
    "rfft2",
    "irfft2",
    "fft2",
    "ifft2",
    "fftn",
    "ifftn",
]

type _NormKind = L["backward", "ortho", "forward"] | None

###

# keep in sync with `ifft`
@overload  # Nd complexfloating
def fft[ShapeT: _Shape, DTypeT: np.dtype[np.complexfloating]](
    a: np.ndarray[ShapeT, DTypeT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd float64 | +integer
def fft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def fft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def fft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +complex
def fft(
    a: Sequence[complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +complex
def fft(
    a: Sequence[Sequence[complex]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d complexfloating
def fft[ScalarT: np.complexfloating](
    a: _ArrayLike[ScalarT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex
def fft(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def fft(
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def fft[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `fft`
@overload  # Nd complexfloating
def ifft[ShapeT: _Shape, DTypeT: np.dtype[np.complexfloating]](
    a: np.ndarray[ShapeT, DTypeT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd float64 | +integer
def ifft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def ifft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def ifft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +complex
def ifft(
    a: Sequence[complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +complex
def ifft(
    a: Sequence[Sequence[complex]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d complexfloating
def ifft[ScalarT: np.complexfloating](
    a: _ArrayLike[ScalarT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex
def ifft(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def ifft(
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def ifft[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `ihfft`
@overload  # Nd float64 | +integer
def rfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def rfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def rfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +float
def rfft(
    a: Sequence[float],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +float
def rfft(
    a: Sequence[Sequence[float]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d +float
def rfft(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], float],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def rfft(
    a: _ArrayLikeFloat_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def rfft[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeFloat_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `hfft`
@overload  # Nd floating
def irfft[ShapeT: _Shape, DTypeT: np.dtype[np.floating]](
    a: np.ndarray[ShapeT, DTypeT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd complex128 | +integer
def irfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex128 | np.integer | np.bool]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float64]]: ...
@overload  # Nd complex64
def irfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex64]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float32]]: ...
@overload  # Nd clongdouble
def irfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.clongdouble]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.longdouble]]: ...
@overload  # 1d +complex
def irfft(
    a: Sequence[complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.float64]]: ...
@overload  # 2d +complex
def irfft(
    a: Sequence[Sequence[complex]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.float64]]: ...
@overload  # ?d floating
def irfft[ScalarT: np.floating](
    a: _ArrayLike[ScalarT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex | complex128 | +integer
def irfft(
    a: _DualArrayLike[np.dtype[np.complex128 | np.integer | np.bool], complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.float64]: ...
@overload  # fallback
def irfft(
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.floating]: ...
@overload  # out: <given>
def irfft[ArrayT: NDArray[np.floating]](
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `irfft` above
@overload  # Nd floating
def hfft[ShapeT: _Shape, DTypeT: np.dtype[np.floating]](
    a: np.ndarray[ShapeT, DTypeT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd complex128 | +integer
def hfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex128 | np.integer | np.bool]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float64]]: ...
@overload  # Nd complex64
def hfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex64]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float32]]: ...
@overload  # Nd clongdouble
def hfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.clongdouble]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.longdouble]]: ...
@overload  # 1d +complex
def hfft(
    a: Sequence[complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.float64]]: ...
@overload  # 2d +complex
def hfft(
    a: Sequence[Sequence[complex]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.float64]]: ...
@overload  # ?d floating
def hfft[ScalarT: np.floating](
    a: _ArrayLike[ScalarT],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex | complex128 | +integer
def hfft(
    a: _DualArrayLike[np.dtype[np.complex128 | np.integer | np.bool], complex],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.float64]: ...
@overload  # fallback
def hfft(
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.floating]: ...
@overload  # out: <given>
def hfft[ArrayT: NDArray[np.floating]](
    a: _ArrayLikeNumber_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `rfft`
@overload  # Nd float64 | +integer
def ihfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def ihfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def ihfft[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +float
def ihfft(
    a: Sequence[float],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +float
def ihfft(
    a: Sequence[Sequence[float]],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d +float
def ihfft(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], float],
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def ihfft(
    a: _ArrayLikeFloat_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def ihfft[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeFloat_co,
    n: int | None = None,
    axis: int = -1,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `ifftn`
@overload  # Nd complexfloating
def fftn[ShapeT: _Shape, DTypeT: np.dtype[np.complexfloating]](
    a: np.ndarray[ShapeT, DTypeT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd float64 | +integer
def fftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def fftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def fftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +complex
def fftn(
    a: Sequence[complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +complex
def fftn(
    a: Sequence[Sequence[complex]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d complexfloating
def fftn[ScalarT: np.complexfloating](
    a: _ArrayLike[ScalarT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex
def fftn(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def fftn(
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def fftn[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `fftn`
@overload  # Nd complexfloating
def ifftn[ShapeT: _Shape, DTypeT: np.dtype[np.complexfloating]](
    a: np.ndarray[ShapeT, DTypeT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd float64 | +integer
def ifftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def ifftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def ifftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +complex
def ifftn(
    a: Sequence[complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +complex
def ifftn(
    a: Sequence[Sequence[complex]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d complexfloating
def ifftn[ScalarT: np.complexfloating](
    a: _ArrayLike[ScalarT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex
def ifftn(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def ifftn(
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def ifftn[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

#
@overload  # Nd float64 | +integer
def rfftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def rfftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def rfftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +float
def rfftn(
    a: Sequence[float],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +float
def rfftn(
    a: Sequence[Sequence[float]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d +float
def rfftn(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], float],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def rfftn(
    a: _ArrayLikeFloat_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def rfftn[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeFloat_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

#
@overload  # Nd floating
def irfftn[ShapeT: _Shape, DTypeT: np.dtype[np.floating]](
    a: np.ndarray[ShapeT, DTypeT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd complex128 | +integer
def irfftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex128 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float64]]: ...
@overload  # Nd complex64
def irfftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex64]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float32]]: ...
@overload  # Nd clongdouble
def irfftn[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.clongdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.longdouble]]: ...
@overload  # 1d +complex
def irfftn(
    a: Sequence[complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.float64]]: ...
@overload  # 2d +complex
def irfftn(
    a: Sequence[Sequence[complex]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.float64]]: ...
@overload  # ?d floating
def irfftn[ScalarT: np.floating](
    a: _ArrayLike[ScalarT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex | complex128 | +integer
def irfftn(
    a: _DualArrayLike[np.dtype[np.complex128 | np.integer | np.bool], complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.float64]: ...
@overload  # fallback
def irfftn(
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.floating]: ...
@overload  # out: <given>
def irfftn[ArrayT: NDArray[np.floating]](
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = None,
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `ifft2`
@overload  # Nd complexfloating
def fft2[ShapeT: _Shape, DTypeT: np.dtype[np.complexfloating]](
    a: np.ndarray[ShapeT, DTypeT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd float64 | +integer
def fft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def fft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def fft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +complex
def fft2(
    a: Sequence[complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +complex
def fft2(
    a: Sequence[Sequence[complex]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d complexfloating
def fft2[ScalarT: np.complexfloating](
    a: _ArrayLike[ScalarT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex
def fft2(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def fft2(
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def fft2[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

# keep in sync with `fft2`
@overload  # Nd complexfloating
def ifft2[ShapeT: _Shape, DTypeT: np.dtype[np.complexfloating]](
    a: np.ndarray[ShapeT, DTypeT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd float64 | +integer
def ifft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def ifft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def ifft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +complex
def ifft2(
    a: Sequence[complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +complex
def ifft2(
    a: Sequence[Sequence[complex]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d complexfloating
def ifft2[ScalarT: np.complexfloating](
    a: _ArrayLike[ScalarT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex
def ifft2(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def ifft2(
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def ifft2[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

#
@overload  # Nd float64 | +integer
def rfft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float64 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex128]]: ...
@overload  # Nd float32 | float16
def rfft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.float32 | np.float16]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.complex64]]: ...
@overload  # Nd longdouble
def rfft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.longdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.clongdouble]]: ...
@overload  # 1d +float
def rfft2(
    a: Sequence[float],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.complex128]]: ...
@overload  # 2d +float
def rfft2(
    a: Sequence[Sequence[float]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.complex128]]: ...
@overload  # ?d +float
def rfft2(
    a: _DualArrayLike[np.dtype[np.float64 | np.integer | np.bool], float],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complex128]: ...
@overload  # fallback
def rfft2(
    a: _ArrayLikeFloat_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.complexfloating]: ...
@overload  # out: <given>
def rfft2[ArrayT: NDArray[np.complexfloating]](
    a: _ArrayLikeFloat_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...

#
@overload  # Nd floating
def irfft2[ShapeT: _Shape, DTypeT: np.dtype[np.floating]](
    a: np.ndarray[ShapeT, DTypeT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, DTypeT]: ...
@overload  # Nd complex128 | +integer
def irfft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex128 | np.integer | np.bool]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float64]]: ...
@overload  # Nd complex64
def irfft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.complex64]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.float32]]: ...
@overload  # Nd clongdouble
def irfft2[ShapeT: _Shape](
    a: np.ndarray[ShapeT, np.dtype[np.clongdouble]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[ShapeT, np.dtype[np.longdouble]]: ...
@overload  # 1d +complex
def irfft2(
    a: Sequence[complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int], np.dtype[np.float64]]: ...
@overload  # 2d +complex
def irfft2(
    a: Sequence[Sequence[complex]],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> np.ndarray[tuple[int, int], np.dtype[np.float64]]: ...
@overload  # ?d floating
def irfft2[ScalarT: np.floating](
    a: _ArrayLike[ScalarT],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[ScalarT]: ...
@overload  # ?d +complex | complex128 | +integer
def irfft2(
    a: _DualArrayLike[np.dtype[np.complex128 | np.integer | np.bool], complex],
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.float64]: ...
@overload  # fallback
def irfft2(
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    out: None = None,
) -> NDArray[np.floating]: ...
@overload  # out: <given>
def irfft2[ArrayT: NDArray[np.floating]](
    a: _ArrayLikeNumber_co,
    s: Sequence[int] | None = None,
    axes: Sequence[int] | None = (-2, -1),
    norm: _NormKind = None,
    *,
    out: ArrayT,
) -> ArrayT: ...



"""
============================
Typing (:mod:`numpy.typing`)
============================

.. versionadded:: 1.20

Large parts of the NumPy API have :pep:`484`-style type annotations. In
addition a number of type aliases are available to users, most prominently
the two below:

- `ArrayLike`: objects that can be converted to arrays
- `DTypeLike`: objects that can be converted to dtypes

.. _typing-extensions: https://pypi.org/project/typing-extensions/

Mypy plugin
-----------

.. versionadded:: 1.21

.. automodule:: numpy.typing.mypy_plugin

.. currentmodule:: numpy.typing

Differences from the runtime NumPy API
--------------------------------------

NumPy is very flexible. Trying to describe the full range of
possibilities statically would result in types that are not very
helpful. For that reason, the typed NumPy API is often stricter than
the runtime NumPy API. This section describes some notable
differences.

ArrayLike
~~~~~~~~~

The `ArrayLike` type tries to avoid creating object arrays. For
example,

.. code-block:: python

    >>> np.array(x**2 for x in range(10))
    array(<generator object <genexpr> at ...>, dtype=object)

is valid NumPy code which will create a 0-dimensional object
array. Type checkers will complain about the above example when using
the NumPy types however. If you really intended to do the above, then
you can either use a ``# type: ignore`` comment:

.. code-block:: python

    >>> np.array(x**2 for x in range(10))  # type: ignore

or explicitly type the array like object as `~typing.Any`:

.. code-block:: python

    >>> from typing import Any
    >>> array_like: Any = (x**2 for x in range(10))
    >>> np.array(array_like)
    array(<generator object <genexpr> at ...>, dtype=object)

DTypeLike
~~~~~~~~~

The `DTypeLike` type tries to avoid creation of dtype objects using
dictionary of fields like below:

.. code-block:: python

    >>> x = np.dtype({"field1": (float, 1), "field2": (int, 3)})

Although this is valid NumPy code, the type checker will complain about it,
since its usage is discouraged.
Please see : :ref:`Data type objects <arrays.dtypes>`

Number precision
~~~~~~~~~~~~~~~~

The precision of `numpy.number` subclasses is treated as an invariant generic
parameter (see :class:`~NBitBase`), simplifying the annotating of processes
involving precision-based casting.

.. code-block:: python

    >>> from typing import TypeVar
    >>> import numpy as np
    >>> import numpy.typing as npt

    >>> T = TypeVar("T", bound=npt.NBitBase)
    >>> def func(a: np.floating[T], b: np.floating[T]) -> np.floating[T]:
    ...     ...

Consequently, the likes of `~numpy.float16`, `~numpy.float32` and
`~numpy.float64` are still sub-types of `~numpy.floating`, but, contrary to
runtime, they're not necessarily considered as sub-classes.

.. deprecated:: 2.3
    The :class:`~numpy.typing.NBitBase` helper is deprecated and will be
    removed in a future release. Prefer expressing precision relationships via
    ``typing.overload`` or ``TypeVar`` definitions bounded by concrete scalar
    classes. For example:

    .. code-block:: python

        from typing import TypeVar
        import numpy as np

        S = TypeVar("S", bound=np.floating)

        def func(a: S, b: S) -> S:
            ...

    or in the case of different input types mapping to different output types:

   .. code-block:: python

        from typing import overload
        import numpy as np

        @overload
        def phase(x: np.complex64) -> np.float32: ...
        @overload
        def phase(x: np.complex128) -> np.float64: ...
        @overload
        def phase(x: np.clongdouble) -> np.longdouble: ...
        def phase(x: np.complexfloating) -> np.floating:
            ...

Timedelta64
~~~~~~~~~~~

The `~numpy.timedelta64` class is not considered a subclass of
`~numpy.signedinteger`, the former only inheriting from `~numpy.generic`
while static type checking.

Record array dtypes
~~~~~~~~~~~~~~~~~~~

The dtype of `numpy.recarray`, and the :ref:`routines.array-creation.rec`
functions in general, can be specified in one of two ways:

* Directly via the ``dtype`` argument.
* With up to five helper arguments that operate via `numpy.rec.format_parser`:
  ``formats``, ``names``, ``titles``, ``aligned`` and ``byteorder``.

These two approaches are currently typed as being mutually exclusive,
*i.e.* if ``dtype`` is specified than one may not specify ``formats``.
While this mutual exclusivity is not (strictly) enforced during runtime,
combining both dtype specifiers can lead to unexpected or even downright
buggy behavior.

API
---

.. rubric:: ndarray

The `numpy.ndarray` class is a `generic type`_ that accepts two type arguments:

1. The type of `numpy.ndarray.shape`, which must be a `tuple` of `int`, e.g.
   ``tuple[int, int]`` (2-D shape) or ``tuple[()]`` (0-D shape).
   The default shape is ``tuple[Any, ...]``, which represents an unknown shape with
   *any* number of dimensions.
   Currently, ``Literal`` ints or other more specific types are not supported.
2. The type of `numpy.ndarray.dtype`, which must be a subtype of `numpy.dtype` such as
   ``numpy.dtype[numpy.float64]``. If omitted, it will default to ``numpy.dtype[Any]``.

.. code-block:: python

    >>> import numpy as np

    >>> type ImageRGB = np.ndarray[tuple[int, int, int], np.dtype[np.uint8]]
    >>> type Vector[S: np.generic] = np.ndarray[tuple[int], np.dtype[S]]

.. _generic type: https://typing.python.org/en/latest/spec/generics.html

"""
# NOTE: The API section will be appended with additional entries
# further down in this file

# pyright: reportDeprecated=false

from numpy._typing import ArrayLike, DTypeLike, NBitBase, NDArray

__all__ = ["ArrayLike", "DTypeLike", "NBitBase", "NDArray"]


__DIR = __all__ + [k for k in globals() if k.startswith("__") and k.endswith("__")]
__DIR_SET = frozenset(__DIR)


def __dir__() -> list[str]:
    return __DIR

def __getattr__(name: str) -> object:
    if name == "NBitBase":
        import warnings

        # Deprecated in NumPy 2.3, 2025-05-01
        warnings.warn(
            "`NBitBase` is deprecated and will be removed from numpy.typing in the "
            "future. Use `@typing.overload` or a `TypeVar` with a scalar-type as upper "
            "bound, instead. (deprecated in NumPy 2.3)",
            DeprecationWarning,
            stacklevel=2,
        )
        return NBitBase

    if name in __DIR_SET:
        return globals()[name]

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


if __doc__ is not None:
    from numpy._typing._add_docstring import _docstrings
    __doc__ += _docstrings
    __doc__ += '\n.. autoclass:: numpy.typing.NBitBase\n'
    del _docstrings

from numpy._pytesttester import PytestTester

test = PytestTester(__name__)
del PytestTester



from pathlib import Path
import shutil

import numpy as np
import pytest
from pytest import approx

from matplotlib import _image
from matplotlib.testing.compare import compare_images
from matplotlib.testing.decorators import _image_directories
from matplotlib.testing.exceptions import ImageComparisonFailure


# Tests of the image comparison algorithm.
@pytest.mark.parametrize(
    'im1, im2, tol, expect_rms',
    [
        # Comparison of an image and the same image with minor differences.
        # This expects the images to compare equal under normal tolerance, and
        # have a small RMS.
        ('basn3p02.png', 'basn3p02-minorchange.png', 10, None),
        # Now test with no tolerance.
        ('basn3p02.png', 'basn3p02-minorchange.png', 0, 6.50646),
        # Comparison with an image that is shifted by 1px in the X axis.
        ('basn3p02.png', 'basn3p02-1px-offset.png', 0, 90.15611),
        # Comparison with an image with half the pixels shifted by 1px in the X
        # axis.
        ('basn3p02.png', 'basn3p02-half-1px-offset.png', 0, 63.75),
        # Comparison of an image and the same image scrambled.
        # This expects the images to compare completely different, with a very
        # large RMS.
        # Note: The image has been scrambled in a specific way, by having
        # each color component of each pixel randomly placed somewhere in the
        # image. It contains exactly the same number of pixels of each color
        # value of R, G and B, but in a totally different position.
        # Test with no tolerance to make sure that we pick up even a very small
        # RMS error.
        ('basn3p02.png', 'basn3p02-scrambled.png', 0, 172.63582),
        # Comparison of an image and a slightly brighter image.
        # The two images are solid color, with the second image being exactly 1
        # color value brighter.
        # This expects the images to compare equal under normal tolerance, and
        # have an RMS of exactly 1.
        ('all127.png', 'all128.png', 0, 1),
        # Now test the reverse comparison.
        ('all128.png', 'all127.png', 0, 1),
    ])
def test_image_comparison_expect_rms(im1, im2, tol, expect_rms, tmp_path,
                                     monkeypatch):
    """
    Compare two images, expecting a particular RMS error.

    im1 and im2 are filenames relative to the baseline_dir directory.

    tol is the tolerance to pass to compare_images.

    expect_rms is the expected RMS value, or None. If None, the test will
    succeed if compare_images succeeds. Otherwise, the test will succeed if
    compare_images fails and returns an RMS error almost equal to this value.
    """
    # Change the working directory using monkeypatch to use a temporary
    # test specific directory
    monkeypatch.chdir(tmp_path)
    baseline_dir, result_dir = map(Path, _image_directories(lambda: "dummy"))
    # Copy "test" image to result_dir, so that compare_images writes
    # the diff to result_dir, rather than to the source tree
    result_im2 = result_dir / im1
    shutil.copyfile(baseline_dir / im2, result_im2)
    results = compare_images(
        baseline_dir / im1, result_im2, tol=tol, in_decorator=True)

    if expect_rms is None:
        assert results is None
    else:
        assert results is not None
        assert results['rms'] == approx(expect_rms, abs=1e-4)


def test_invalid_input():
    img = np.zeros((16, 16, 4), dtype=np.uint8)

    with pytest.raises(ImageComparisonFailure,
                       match='must be 3-dimensional, but is 2-dimensional'):
        _image.calculate_rms_and_diff(img[:, :, 0], img)
    with pytest.raises(ImageComparisonFailure,
                       match='must be 3-dimensional, but is 5-dimensional'):
        _image.calculate_rms_and_diff(img, img[:, :, :, np.newaxis, np.newaxis])
    with pytest.raises(ImageComparisonFailure,
                       match='must be RGB or RGBA but has depth 2'):
        _image.calculate_rms_and_diff(img[:, :, :2], img)

    with pytest.raises(ImageComparisonFailure,
                       match=r'expected size: \(16, 16, 4\) actual size \(8, 16, 4\)'):
        _image.calculate_rms_and_diff(img, img[:8, :, :])
    with pytest.raises(ImageComparisonFailure,
                       match=r'expected size: \(16, 16, 4\) actual size \(16, 6, 4\)'):
        _image.calculate_rms_and_diff(img, img[:, :6, :])
    with pytest.raises(ImageComparisonFailure,
                       match=r'expected size: \(16, 16, 4\) actual size \(16, 16, 3\)'):
        _image.calculate_rms_and_diff(img, img[:, :, :3])



"""Run binary A/B tests from aggregate count data.

This module validates inputs, sets up the random number generator so results are
reproducible from a seed, and combines the frequentist and Bayesian helpers to
build a :class:BinaryABResult.
"""

from __future__ import annotations

import math
from dataclasses import dataclass

import numpy as np

from . import bayesian, frequentist, validation
from .result import BinaryABResult


@dataclass(frozen=True)
class BinaryABTest:
    """A binary A/B test using aggregate success and total counts."""

    control_successes: int
    control_total: int
    treatment_successes: int
    treatment_total: int
    prior_alpha: float
    prior_beta: float
    n_simulations: int
    credible_interval: float
    seed: int | None

    @classmethod
    def from_counts(
        cls,
        control_successes: int,
        control_total: int,
        treatment_successes: int,
        treatment_total: int,
        *,
        prior_alpha: float = 1.0,
        prior_beta: float = 1.0,
        n_simulations: int = 100_000,
        credible_interval: float = 0.95,
        seed: int | None = None,
    ) -> "BinaryABTest":
        """Create a binary A/B test from aggregate counts.

        The four count inputs can be positional. All other settings have to be names.
        The default prior is Beta(1, 1).
        """
        validation.validate_count("control_successes", control_successes)
        validation.validate_total("control_total", control_total)
        validation.validate_count("treatment_successes", treatment_successes)
        validation.validate_total("treatment_total", treatment_total)
        validation.validate_successes_le_total(
            "control_successes", control_successes, "control_total", control_total
        )
        validation.validate_successes_le_total(
            "treatment_successes",
            treatment_successes,
            "treatment_total",
            treatment_total,
        )
        validation.validate_prior("prior_alpha", prior_alpha)
        validation.validate_prior("prior_beta", prior_beta)
        validation.validate_n_simulations(n_simulations)
        validation.validate_credible_interval(credible_interval)
        validation.validate_seed(seed)

        return cls(
            control_successes=control_successes,
            control_total=control_total,
            treatment_successes=treatment_successes,
            treatment_total=treatment_total,
            prior_alpha=float(prior_alpha),
            prior_beta=float(prior_beta),
            n_simulations=n_simulations,
            credible_interval=credible_interval,
            seed=seed,
        )

    def run(self) -> BinaryABResult:
        """Run the frequentist and Bayesian analyses and return the result."""
        control_rate = self.control_successes / self.control_total
        treatment_rate = self.treatment_successes / self.treatment_total
        absolute_lift = treatment_rate - control_rate

        # Relative lift is undefined when the control rate is zero.
        relative_lift = (
            absolute_lift / control_rate if control_rate != 0 else math.nan
        )

        # --- Frequentist ---
        z_statistic, p_value = frequentist.two_proportion_z_test(
            self.control_successes,
            self.control_total,
            self.treatment_successes,
            self.treatment_total,
        )

        # --- Bayesian ---
        rng = np.random.default_rng(self.seed)
        lift_samples = bayesian.simulate_lift_samples(
            self.control_successes,
            self.control_total,
            self.treatment_successes,
            self.treatment_total,
            self.prior_alpha,
            self.prior_beta,
            self.n_simulations,
            rng,
        )

        lower, upper = bayesian.credible_interval_bounds(
            lift_samples, self.credible_interval
        )

        return BinaryABResult(
            control_successes=self.control_successes,
            control_total=self.control_total,
            treatment_successes=self.treatment_successes,
            treatment_total=self.treatment_total,
            prior_alpha=self.prior_alpha,
            prior_beta=self.prior_beta,
            n_simulations=self.n_simulations,
            credible_interval=self.credible_interval,
            seed=self.seed,
            control_rate=control_rate,
            treatment_rate=treatment_rate,
            absolute_lift=absolute_lift,
            relative_lift=relative_lift,
            z_statistic=z_statistic,
            p_value=p_value,
            posterior_mean_lift=float(np.mean(lift_samples)),
            posterior_median_lift=float(np.median(lift_samples)),
            prob_treatment_better=float(np.mean(lift_samples > 0)),
            prob_control_better=float(np.mean(lift_samples < 0)),
            credible_interval_bounds=(lower, upper),
            expected_loss_treatment=bayesian.expected_loss_treatment(lift_samples),
            expected_loss_control=bayesian.expected_loss_control(lift_samples),
            lift_samples=lift_samples,
        )



import pytest

import numpy as np
from numpy._core._rational_tests import rational, rational2
from numpy.lib._stride_tricks_impl import (
    _broadcast_shape,
    as_strided,
    broadcast_arrays,
    broadcast_shapes,
    broadcast_to,
    sliding_window_view,
)
from numpy.testing import (
    assert_,
    assert_array_equal,
    assert_equal,
    assert_raises,
    assert_raises_regex,
)


def assert_shapes_correct(input_shapes, expected_shape):
    # Broadcast a list of arrays with the given input shapes and check the
    # common output shape.

    inarrays = [np.zeros(s) for s in input_shapes]
    outarrays = broadcast_arrays(*inarrays)
    outshapes = [a.shape for a in outarrays]
    expected = [expected_shape] * len(inarrays)
    assert_equal(outshapes, expected)


def assert_incompatible_shapes_raise(input_shapes):
    # Broadcast a list of arrays with the given (incompatible) input shapes
    # and check that they raise a ValueError.

    inarrays = [np.zeros(s) for s in input_shapes]
    assert_raises(ValueError, broadcast_arrays, *inarrays)


def assert_same_as_ufunc(shape0, shape1, transposed=False, flipped=False):
    # Broadcast two shapes against each other and check that the data layout
    # is the same as if a ufunc did the broadcasting.

    x0 = np.zeros(shape0, dtype=int)
    # Note that multiply.reduce's identity element is 1.0, so when shape1==(),
    # this gives the desired n==1.
    n = int(np.multiply.reduce(shape1))
    x1 = np.arange(n).reshape(shape1)
    if transposed:
        x0 = x0.T
        x1 = x1.T
    if flipped:
        x0 = x0[::-1]
        x1 = x1[::-1]
    # Use the add ufunc to do the broadcasting. Since we're adding 0s to x1, the
    # result should be exactly the same as the broadcasted view of x1.
    y = x0 + x1
    b0, b1 = broadcast_arrays(x0, x1)
    assert_array_equal(y, b1)


def test_same():
    x = np.arange(10)
    y = np.arange(10)
    bx, by = broadcast_arrays(x, y)
    assert_array_equal(x, bx)
    assert_array_equal(y, by)

def test_broadcast_kwargs():
    # ensure that a TypeError is appropriately raised when
    # np.broadcast_arrays() is called with any keyword
    # argument other than 'subok'
    x = np.arange(10)
    y = np.arange(10)

    with assert_raises_regex(TypeError, 'got an unexpected keyword'):
        broadcast_arrays(x, y, dtype='float64')


def test_one_off():
    x = np.array([[1, 2, 3]])
    y = np.array([[1], [2], [3]])
    bx, by = broadcast_arrays(x, y)
    bx0 = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
    by0 = bx0.T
    assert_array_equal(bx0, bx)
    assert_array_equal(by0, by)


def test_same_input_shapes():
    # Check that the final shape is just the input shape.

    data = [
        (),
        (1,),
        (3,),
        (0, 1),
        (0, 3),
        (1, 0),
        (3, 0),
        (1, 3),
        (3, 1),
        (3, 3),
    ]
    for shape in data:
        input_shapes = [shape]
        # Single input.
        assert_shapes_correct(input_shapes, shape)
        # Double input.
        input_shapes2 = [shape, shape]
        assert_shapes_correct(input_shapes2, shape)
        # Triple input.
        input_shapes3 = [shape, shape, shape]
        assert_shapes_correct(input_shapes3, shape)


def test_two_compatible_by_ones_input_shapes():
    # Check that two different input shapes of the same length, but some have
    # ones, broadcast to the correct shape.

    data = [
        [[(1,), (3,)], (3,)],
        [[(1, 3), (3, 3)], (3, 3)],
        [[(3, 1), (3, 3)], (3, 3)],
        [[(1, 3), (3, 1)], (3, 3)],
        [[(1, 1), (3, 3)], (3, 3)],
        [[(1, 1), (1, 3)], (1, 3)],
        [[(1, 1), (3, 1)], (3, 1)],
        [[(1, 0), (0, 0)], (0, 0)],
        [[(0, 1), (0, 0)], (0, 0)],
        [[(1, 0), (0, 1)], (0, 0)],
        [[(1, 1), (0, 0)], (0, 0)],
        [[(1, 1), (1, 0)], (1, 0)],
        [[(1, 1), (0, 1)], (0, 1)],
    ]
    for input_shapes, expected_shape in data:
        assert_shapes_correct(input_shapes, expected_shape)
        # Reverse the input shapes since broadcasting should be symmetric.
        assert_shapes_correct(input_shapes[::-1], expected_shape)


def test_two_compatible_by_prepending_ones_input_shapes():
    # Check that two different input shapes (of different lengths) broadcast
    # to the correct shape.

    data = [
        [[(), (3,)], (3,)],
        [[(3,), (3, 3)], (3, 3)],
        [[(3,), (3, 1)], (3, 3)],
        [[(1,), (3, 3)], (3, 3)],
        [[(), (3, 3)], (3, 3)],
        [[(1, 1), (3,)], (1, 3)],
        [[(1,), (3, 1)], (3, 1)],
        [[(1,), (1, 3)], (1, 3)],
        [[(), (1, 3)], (1, 3)],
        [[(), (3, 1)], (3, 1)],
        [[(), (0,)], (0,)],
        [[(0,), (0, 0)], (0, 0)],
        [[(0,), (0, 1)], (0, 0)],
        [[(1,), (0, 0)], (0, 0)],
        [[(), (0, 0)], (0, 0)],
        [[(1, 1), (0,)], (1, 0)],
        [[(1,), (0, 1)], (0, 1)],
        [[(1,), (1, 0)], (1, 0)],
        [[(), (1, 0)], (1, 0)],
        [[(), (0, 1)], (0, 1)],
    ]
    for input_shapes, expected_shape in data:
        assert_shapes_correct(input_shapes, expected_shape)
        # Reverse the input shapes since broadcasting should be symmetric.
        assert_shapes_correct(input_shapes[::-1], expected_shape)


def test_incompatible_shapes_raise_valueerror():
    # Check that a ValueError is raised for incompatible shapes.

    data = [
        [(3,), (4,)],
        [(2, 3), (2,)],
        [(3,), (3,), (4,)],
        [(1, 3, 4), (2, 3, 3)],
    ]
    for input_shapes in data:
        assert_incompatible_shapes_raise(input_shapes)
        # Reverse the input shapes since broadcasting should be symmetric.
        assert_incompatible_shapes_raise(input_shapes[::-1])


def test_same_as_ufunc():
    # Check that the data layout is the same as if a ufunc did the operation.

    data = [
        [[(1,), (3,)], (3,)],
        [[(1, 3), (3, 3)], (3, 3)],
        [[(3, 1), (3, 3)], (3, 3)],
        [[(1, 3), (3, 1)], (3, 3)],
        [[(1, 1), (3, 3)], (3, 3)],
        [[(1, 1), (1, 3)], (1, 3)],
        [[(1, 1), (3, 1)], (3, 1)],
        [[(1, 0), (0, 0)], (0, 0)],
        [[(0, 1), (0, 0)], (0, 0)],
        [[(1, 0), (0, 1)], (0, 0)],
        [[(1, 1), (0, 0)], (0, 0)],
        [[(1, 1), (1, 0)], (1, 0)],
        [[(1, 1), (0, 1)], (0, 1)],
        [[(), (3,)], (3,)],
        [[(3,), (3, 3)], (3, 3)],
        [[(3,), (3, 1)], (3, 3)],
        [[(1,), (3, 3)], (3, 3)],
        [[(), (3, 3)], (3, 3)],
        [[(1, 1), (3,)], (1, 3)],
        [[(1,), (3, 1)], (3, 1)],
        [[(1,), (1, 3)], (1, 3)],
        [[(), (1, 3)], (1, 3)],
        [[(), (3, 1)], (3, 1)],
        [[(), (0,)], (0,)],
        [[(0,), (0, 0)], (0, 0)],
        [[(0,), (0, 1)], (0, 0)],
        [[(1,), (0, 0)], (0, 0)],
        [[(), (0, 0)], (0, 0)],
        [[(1, 1), (0,)], (1, 0)],
        [[(1,), (0, 1)], (0, 1)],
        [[(1,), (1, 0)], (1, 0)],
        [[(), (1, 0)], (1, 0)],
        [[(), (0, 1)], (0, 1)],
    ]
    for input_shapes, expected_shape in data:
        assert_same_as_ufunc(input_shapes[0], input_shapes[1],
                             f"Shapes: {input_shapes[0]} {input_shapes[1]}")
        # Reverse the input shapes since broadcasting should be symmetric.
        assert_same_as_ufunc(input_shapes[1], input_shapes[0])
        # Try them transposed, too.
        assert_same_as_ufunc(input_shapes[0], input_shapes[1], True)
        # ... and flipped for non-rank-0 inputs in order to test negative
        # strides.
        if () not in input_shapes:
            assert_same_as_ufunc(input_shapes[0], input_shapes[1], False, True)
            assert_same_as_ufunc(input_shapes[0], input_shapes[1], True, True)


def test_broadcast_to_succeeds():
    data = [
        [np.array(0), (0,), np.array(0)],
        [np.array(0), (1,), np.zeros(1)],
        [np.array(0), (3,), np.zeros(3)],
        [np.ones(1), (1,), np.ones(1)],
        [np.ones(1), (2,), np.ones(2)],
        [np.ones(1), (1, 2, 3), np.ones((1, 2, 3))],
        [np.arange(3), (3,), np.arange(3)],
        [np.arange(3), (1, 3), np.arange(3).reshape(1, -1)],
        [np.arange(3), (2, 3), np.array([[0, 1, 2], [0, 1, 2]])],
        # test if shape is not a tuple
        [np.ones(0), 0, np.ones(0)],
        [np.ones(1), 1, np.ones(1)],
        [np.ones(1), 2, np.ones(2)],
        # these cases with size 0 are strange, but they reproduce the behavior
        # of broadcasting with ufuncs (see test_same_as_ufunc above)
        [np.ones(1), (0,), np.ones(0)],
        [np.ones((1, 2)), (0, 2), np.ones((0, 2))],
        [np.ones((2, 1)), (2, 0), np.ones((2, 0))],
    ]
    for input_array, shape, expected in data:
        actual = broadcast_to(input_array, shape)
        assert_array_equal(expected, actual)


def test_broadcast_to_raises():
    data = [
        [(0,), ()],
        [(1,), ()],
        [(3,), ()],
        [(3,), (1,)],
        [(3,), (2,)],
        [(3,), (4,)],
        [(1, 2), (2, 1)],
        [(1, 1), (1,)],
        [(1,), -1],
        [(1,), (-1,)],
        [(1, 2), (-1, 2)],
    ]
    for orig_shape, target_shape in data:
        arr = np.zeros(orig_shape)
        assert_raises(ValueError, lambda: broadcast_to(arr, target_shape))


def test_broadcast_shape():
    # tests internal _broadcast_shape
    # _broadcast_shape is already exercised indirectly by broadcast_arrays
    # _broadcast_shape is also exercised by the public broadcast_shapes function
    assert_equal(_broadcast_shape(), ())
    assert_equal(_broadcast_shape([1, 2]), (2,))
    assert_equal(_broadcast_shape(np.ones((1, 1))), (1, 1))
    assert_equal(_broadcast_shape(np.ones((1, 1)), np.ones((3, 4))), (3, 4))
    assert_equal(_broadcast_shape(*([np.ones((1, 2))] * 32)), (1, 2))
    assert_equal(_broadcast_shape(*([np.ones((1, 2))] * 100)), (1, 2))

    # regression tests for gh-5862
    assert_equal(_broadcast_shape(*([np.ones(2)] * 32 + [1])), (2,))
    bad_args = [np.ones(2)] * 32 + [np.ones(3)] * 32
    assert_raises(ValueError, lambda: _broadcast_shape(*bad_args))


def test_broadcast_shapes_succeeds():
    # tests public broadcast_shapes
    data = [
        [[], ()],
        [[()], ()],
        [[(7,)], (7,)],
        [[(1, 2), (2,)], (1, 2)],
        [[(1, 1)], (1, 1)],
        [[(1, 1), (3, 4)], (3, 4)],
        [[(6, 7), (5, 6, 1), (7,), (5, 1, 7)], (5, 6, 7)],
        [[(5, 6, 1)], (5, 6, 1)],
        [[(1, 3), (3, 1)], (3, 3)],
        [[(1, 0), (0, 0)], (0, 0)],
        [[(0, 1), (0, 0)], (0, 0)],
        [[(1, 0), (0, 1)], (0, 0)],
        [[(1, 1), (0, 0)], (0, 0)],
        [[(1, 1), (1, 0)], (1, 0)],
        [[(1, 1), (0, 1)], (0, 1)],
        [[(), (0,)], (0,)],
        [[(0,), (0, 0)], (0, 0)],
        [[(0,), (0, 1)], (0, 0)],
        [[(1,), (0, 0)], (0, 0)],
        [[(), (0, 0)], (0, 0)],
        [[(1, 1), (0,)], (1, 0)],
        [[(1,), (0, 1)], (0, 1)],
        [[(1,), (1, 0)], (1, 0)],
        [[(), (1, 0)], (1, 0)],
        [[(), (0, 1)], (0, 1)],
        [[(1,), (3,)], (3,)],
        [[2, (3, 2)], (3, 2)],
    ]
    for input_shapes, target_shape in data:
        assert_equal(broadcast_shapes(*input_shapes), target_shape)

    assert_equal(broadcast_shapes(*([(1, 2)] * 32)), (1, 2))
    assert_equal(broadcast_shapes(*([(1, 2)] * 100)), (1, 2))

    # regression tests for gh-5862
    assert_equal(broadcast_shapes(*([(2,)] * 32)), (2,))


def test_broadcast_shapes_raises():
    # tests public broadcast_shapes
    data = [
        [(3,), (4,)],
        [(2, 3), (2,)],
        [(3,), (3,), (4,)],
        [(1, 3, 4), (2, 3, 3)],
        [(1, 2), (3, 1), (3, 2), (10, 5)],
        [2, (2, 3)],
    ]
    for input_shapes in data:
        assert_raises(ValueError, lambda: broadcast_shapes(*input_shapes))

    bad_args = [(2,)] * 32 + [(3,)] * 32
    assert_raises(ValueError, lambda: broadcast_shapes(*bad_args))


def test_as_strided():
    a = np.array([None])
    a_view = as_strided(a)
    expected = np.array([None])
    assert_array_equal(a_view, np.array([None]))

    a = np.array([1, 2, 3, 4])
    a_view = as_strided(a, shape=(2,), strides=(2 * a.itemsize,))
    expected = np.array([1, 3])
    assert_array_equal(a_view, expected)

    a = np.array([1, 2, 3, 4])
    a_view = as_strided(a, shape=(3, 4), strides=(0, 1 * a.itemsize))
    expected = np.array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]])
    assert_array_equal(a_view, expected)

    # Regression test for gh-5081
    dt = np.dtype([('num', 'i4'), ('obj', 'O')])
    a = np.empty((4,), dtype=dt)
    a['num'] = np.arange(1, 5)
    a_view = as_strided(a, shape=(3, 4), strides=(0, a.itemsize))
    expected_num = [[1, 2, 3, 4]] * 3
    expected_obj = [[None] * 4] * 3
    assert_equal(a_view.dtype, dt)
    assert_array_equal(expected_num, a_view['num'])
    assert_array_equal(expected_obj, a_view['obj'])

    # Make sure that void types without fields are kept unchanged
    a = np.empty((4,), dtype='V4')
    a_view = as_strided(a, shape=(3, 4), strides=(0, a.itemsize))
    assert_equal(a.dtype, a_view.dtype)

    # Make sure that the only type that could fail is properly handled
    dt = np.dtype({'names': [''], 'formats': ['V4']})
    a = np.empty((4,), dtype=dt)
    a_view = as_strided(a, shape=(3, 4), strides=(0, a.itemsize))
    assert_equal(a.dtype, a_view.dtype)

    # Custom dtypes should not be lost (gh-9161)
    r = [rational(i) for i in range(4)]
    a = np.array(r, dtype=rational)
    a_view = as_strided(a, shape=(3, 4), strides=(0, a.itemsize))
    assert_equal(a.dtype, a_view.dtype)
    assert_array_equal([r] * 3, a_view)

    # Also exercise the new-DType-API rational variant.
    r = [rational2(i) for i in range(4)]
    a = np.array(r, dtype=rational2)
    a_view = as_strided(a, shape=(3, 4), strides=(0, a.itemsize))
    assert_equal(a.dtype, a_view.dtype)
    assert_array_equal([r] * 3, a_view)


class TestSlidingWindowView:
    def test_1d(self):
        arr = np.arange(5)
        arr_view = sliding_window_view(arr, 2)
        expected = np.array([[0, 1],
                             [1, 2],
                             [2, 3],
                             [3, 4]])
        assert_array_equal(arr_view, expected)

    def test_2d(self):
        i, j = np.ogrid[:3, :4]
        arr = 10 * i + j
        shape = (2, 2)
        arr_view = sliding_window_view(arr, shape)
        expected = np.array([[[[0, 1], [10, 11]],
                              [[1, 2], [11, 12]],
                              [[2, 3], [12, 13]]],
                             [[[10, 11], [20, 21]],
                              [[11, 12], [21, 22]],
                              [[12, 13], [22, 23]]]])
        assert_array_equal(arr_view, expected)

    def test_2d_with_axis(self):
        i, j = np.ogrid[:3, :4]
        arr = 10 * i + j
        arr_view = sliding_window_view(arr, 3, 0)
        expected = np.array([[[0, 10, 20],
                              [1, 11, 21],
                              [2, 12, 22],
                              [3, 13, 23]]])
        assert_array_equal(arr_view, expected)

    def test_2d_repeated_axis(self):
        i, j = np.ogrid[:3, :4]
        arr = 10 * i + j
        arr_view = sliding_window_view(arr, (2, 3), (1, 1))
        expected = np.array([[[[0, 1, 2],
                               [1, 2, 3]]],
                             [[[10, 11, 12],
                               [11, 12, 13]]],
                             [[[20, 21, 22],
                               [21, 22, 23]]]])
        assert_array_equal(arr_view, expected)

    def test_2d_without_axis(self):
        i, j = np.ogrid[:4, :4]
        arr = 10 * i + j
        shape = (2, 3)
        arr_view = sliding_window_view(arr, shape)
        expected = np.array([[[[0, 1, 2], [10, 11, 12]],
                              [[1, 2, 3], [11, 12, 13]]],
                             [[[10, 11, 12], [20, 21, 22]],
                              [[11, 12, 13], [21, 22, 23]]],
                             [[[20, 21, 22], [30, 31, 32]],
                              [[21, 22, 23], [31, 32, 33]]]])
        assert_array_equal(arr_view, expected)

    def test_errors(self):
        i, j = np.ogrid[:4, :4]
        arr = 10 * i + j
        with pytest.raises(ValueError, match='cannot contain negative values'):
            sliding_window_view(arr, (-1, 3))
        with pytest.raises(
                ValueError,
                match='must provide window_shape for all dimensions of `x`'):
            sliding_window_view(arr, (1,))
        with pytest.raises(
                ValueError,
                match='Must provide matching length window_shape and axis'):
            sliding_window_view(arr, (1, 3, 4), axis=(0, 1))
        with pytest.raises(
                ValueError,
                match='window shape cannot be larger than input array'):
            sliding_window_view(arr, (5, 5))

    def test_writeable(self):
        arr = np.arange(5)
        view = sliding_window_view(arr, 2, writeable=False)
        assert_(not view.flags.writeable)
        with pytest.raises(
                ValueError,
                match='assignment destination is read-only'):
            view[0, 0] = 3
        view = sliding_window_view(arr, 2, writeable=True)
        assert_(view.flags.writeable)
        view[0, 1] = 3
        assert_array_equal(arr, np.array([0, 3, 2, 3, 4]))

    def test_subok(self):
        class MyArray(np.ndarray):
            pass

        arr = np.arange(5).view(MyArray)
        assert_(not isinstance(sliding_window_view(arr, 2,
                                                   subok=False),
                               MyArray))
        assert_(isinstance(sliding_window_view(arr, 2, subok=True), MyArray))
        # Default behavior
        assert_(not isinstance(sliding_window_view(arr, 2), MyArray))


def as_strided_writeable():
    arr = np.ones(10)
    view = as_strided(arr, writeable=False)
    assert_(not view.flags.writeable)

    # Check that writeable also is fine:
    view = as_strided(arr, writeable=True)
    assert_(view.flags.writeable)
    view[...] = 3
    assert_array_equal(arr, np.full_like(arr, 3))

    # Test that things do not break down for readonly:
    arr.flags.writeable = False
    view = as_strided(arr, writeable=False)
    view = as_strided(arr, writeable=True)
    assert_(not view.flags.writeable)


class VerySimpleSubClass(np.ndarray):
    def __new__(cls, *args, **kwargs):
        return np.array(*args, subok=True, **kwargs).view(cls)


class SimpleSubClass(VerySimpleSubClass):
    def __new__(cls, *args, **kwargs):
        self = np.array(*args, subok=True, **kwargs).view(cls)
        self.info = 'simple'
        return self

    def __array_finalize__(self, obj):
        self.info = getattr(obj, 'info', '') + ' finalized'


def test_subclasses():
    # test that subclass is preserved only if subok=True
    a = VerySimpleSubClass([1, 2, 3, 4])
    assert_(type(a) is VerySimpleSubClass)
    a_view = as_strided(a, shape=(2,), strides=(2 * a.itemsize,))
    assert_(type(a_view) is np.ndarray)
    a_view = as_strided(a, shape=(2,), strides=(2 * a.itemsize,), subok=True)
    assert_(type(a_view) is VerySimpleSubClass)
    # test that if a subclass has __array_finalize__, it is used
    a = SimpleSubClass([1, 2, 3, 4])
    a_view = as_strided(a, shape=(2,), strides=(2 * a.itemsize,), subok=True)
    assert_(type(a_view) is SimpleSubClass)
    assert_(a_view.info == 'simple finalized')

    # similar tests for broadcast_arrays
    b = np.arange(len(a)).reshape(-1, 1)
    a_view, b_view = broadcast_arrays(a, b)
    assert_(type(a_view) is np.ndarray)
    assert_(type(b_view) is np.ndarray)
    assert_(a_view.shape == b_view.shape)
    a_view, b_view = broadcast_arrays(a, b, subok=True)
    assert_(type(a_view) is SimpleSubClass)
    assert_(a_view.info == 'simple finalized')
    assert_(type(b_view) is np.ndarray)
    assert_(a_view.shape == b_view.shape)

    # and for broadcast_to
    shape = (2, 4)
    a_view = broadcast_to(a, shape)
    assert_(type(a_view) is np.ndarray)
    assert_(a_view.shape == shape)
    a_view = broadcast_to(a, shape, subok=True)
    assert_(type(a_view) is SimpleSubClass)
    assert_(a_view.info == 'simple finalized')
    assert_(a_view.shape == shape)


def test_writeable():
    # broadcast_to should return a readonly array
    original = np.array([1, 2, 3])
    result = broadcast_to(original, (2, 3))
    assert_equal(result.flags.writeable, False)
    assert_raises(ValueError, result.__setitem__, slice(None), 0)

    # but the result of broadcast_arrays needs to be writeable, to
    # preserve backwards compatibility
    test_cases = [((False,), broadcast_arrays(original,)),
                  ((True, False), broadcast_arrays(0, original))]
    for is_broadcast, results in test_cases:
        for array_is_broadcast, result in zip(is_broadcast, results):
            # This will change to False in a future version
            if array_is_broadcast:
                with pytest.warns(FutureWarning):
                    assert_equal(result.flags.writeable, True)
                with pytest.warns(DeprecationWarning):
                    result[:] = 0
                # Warning not emitted, writing to the array resets it
                assert_equal(result.flags.writeable, True)
            else:
                # No warning:
                assert_equal(result.flags.writeable, True)

    for results in [broadcast_arrays(original),
                    broadcast_arrays(0, original)]:
        for result in results:
            # resets the warn_on_write DeprecationWarning
            result.flags.writeable = True
            # check: no warning emitted
            assert_equal(result.flags.writeable, True)
            result[:] = 0

    # keep readonly input readonly
    original.flags.writeable = False
    _, result = broadcast_arrays(0, original)
    assert_equal(result.flags.writeable, False)

    # regression test for GH6491
    shape = (2,)
    strides = [0]
    tricky_array = as_strided(np.array(0), shape, strides)
    other = np.zeros((1,))
    first, second = broadcast_arrays(tricky_array, other)
    assert_(first.shape == second.shape)


def test_writeable_memoryview():
    # The result of broadcast_arrays exports as a non-writeable memoryview
    # because otherwise there is no good way to opt in to the new behaviour
    # (i.e. you would need to set writeable to False explicitly).
    # See gh-13929.
    original = np.array([1, 2, 3])

    test_cases = [((False, ), broadcast_arrays(original,)),
                  ((True, False), broadcast_arrays(0, original))]
    for is_broadcast, results in test_cases:
        for array_is_broadcast, result in zip(is_broadcast, results):
            # This will change to False in a future version
            if array_is_broadcast:
                # memoryview(result, writable=True) will give warning but cannot
                # be tested using the python API.
                assert memoryview(result).readonly
            else:
                assert not memoryview(result).readonly


def test_reference_types():
    input_array = np.array('a', dtype=object)
    expected = np.array(['a'] * 3, dtype=object)
    actual = broadcast_to(input_array, (3,))
    assert_array_equal(expected, actual)

    actual, _ = broadcast_arrays(input_array, np.ones(3))
    assert_array_equal(expected, actual)


@pytest.mark.parametrize(
    "dtype",
    [
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
        np.float32,
        np.float64,
        np.complex64,
        np.complex128,
    ],
)
def test_as_strided_checked_different_dtypes(dtype):
    """Test as_strided with check_bounds=True with different dtypes."""
    x = np.arange(10, dtype=dtype)
    y = as_strided(x, shape=(5,), strides=(x.itemsize * 2,), check_bounds=True)
    assert y.shape == (5,)
    assert y.dtype == dtype


@pytest.mark.parametrize(
    "size,view_size,stride_mult",
    [
        (10, 5, 1),  # Contiguous view
        (10, 5, 2),  # Every other element
        (20, 10, 2),  # Every other element
        (100, 10, 10),  # Every 10th element
    ],
)
def test_as_strided_checked_1d_positive_strides(size, view_size, stride_mult):
    """Test 1D arrays with positive strides."""
    x = np.arange(size, dtype=np.int64)
    itemsize = x.itemsize
    y = as_strided(
        x, shape=(view_size,), strides=(itemsize * stride_mult,), check_bounds=True
    )
    assert y.shape == (view_size,)
    # Verify data correctness
    expected = x[::stride_mult][:view_size]
    assert_array_equal(y, expected)


@pytest.mark.parametrize(
    "shape,window_shape",
    [
        ((10,), (3,)),
        ((20,), (5,)),
        ((100,), (10,)),
    ],
)
def test_as_strided_checked_sliding_window_1d(shape, window_shape):
    """Test sliding window views in 1D."""
    x = np.arange(shape[0], dtype=np.int64)
    itemsize = x.itemsize
    n_windows = shape[0] - window_shape[0] + 1
    view_shape = (n_windows, window_shape[0])
    view_strides = (itemsize, itemsize)

    y = as_strided(x, shape=view_shape, strides=view_strides, check_bounds=True)
    assert y.shape == view_shape
    # Check first and last windows
    assert_array_equal(y[0], x[: window_shape[0]])
    assert_array_equal(y[-1], x[-window_shape[0] :])


@pytest.mark.parametrize(
    "shape",
    [
        (3, 4),
        (5, 6),
        (10, 10),
    ],
)
def test_as_strided_checked_2d_default_strides(shape):
    """Test 2D arrays with default strides."""
    x = np.arange(np.prod(shape), dtype=np.int64).reshape(shape)
    y = as_strided(x, check_bounds=True)  # Should use default shape and strides
    assert_array_equal(y, x)


@pytest.mark.parametrize("size", [0, 1, 2, 10, 100])
def test_as_strided_checked_zero_stride_broadcasting(size):
    """Test zero strides (broadcasting a single value)."""
    x = np.array([42], dtype=np.int64)
    y = as_strided(x, shape=(size,), strides=(0,), check_bounds=True)
    assert y.shape == (size,)
    if size > 0:
        assert_(np.all(y == 42))


@pytest.mark.parametrize(
    "size,shape,strides",
    [
        # Strides too large
        (10, (5,), (32,)),
        (10, (10,), (16,)),
        (20, (15,), (16,)),
        # Shape too large for strides
        (10, (20,), (8,)),
        (10, (100,), (8,)),
        # 2D out of bounds cases
        (20, (5, 5), (80, 8)),
        (20, (3, 10), (64, 8)),
        # Negative strides that go before array start
        (10, (5,), (-8,)),
        (10, (10,), (-8,)),
        (20, (5,), (-16,)),
        # ND negative strides
        (10, (2, 3, 4), (96, 32, -8)),
        (20, (3, 4), (64, -8)),
        (30, (2, 3, 4), (-96, 32, 8)),
    ],
)
def test_as_strided_checked_out_of_bounds_positive_strides(size, shape, strides):
    """Test that out-of-bounds positive strides raise ValueError."""
    x = np.arange(size, dtype=np.int64)
    with pytest.raises(ValueError, match="out of bounds"):
        as_strided(x, shape=shape, strides=strides, check_bounds=True)


def test_as_strided_checked_view_of_larger_array():
    """Test as_strided

    - with check_bounds=True
    - considers the base array bounds, not just the view.

    """
    a = np.arange(1000, dtype=np.int64)

    b = a[:2]

    # This should succeed because the underlying array has enough memory
    y = as_strided(b, shape=(2,), strides=(400,), check_bounds=True)
    assert_equal(y.shape, (2,))
    assert_equal(y[0], 0)
    assert_equal(y[1], 50)


def test_as_strided_checked_view_with_offset():
    """Test as_strided

    - with check_bounds=True
    - on a view that doesn't start at the beginning.
    """
    a = np.arange(1000, dtype=np.int64)

    b = a[100:102]

    y = as_strided(b, shape=(2,), strides=(80,), check_bounds=True)
    assert_equal(y.shape, (2,))
    assert_equal(y[0], 100)
    assert_equal(y[1], 110)


def test_as_strided_checked_view_out_of_bounds_negative():
    """Test that negative strides on a view correctly detect out of bounds."""
    a = np.arange(1000, dtype=np.int64)

    b = a[5:7]

    with pytest.raises(ValueError, match="out of bounds"):
        as_strided(b, shape=(2,), strides=(-48,), check_bounds=True)


def test_as_strided_checked_view_out_of_bounds_positive():
    """Test that positive strides on a view correctly detect out of bounds."""
    a = np.arange(100, dtype=np.int64)

    b = a[95:97]

    with pytest.raises(ValueError, match="out of bounds"):
        as_strided(b, shape=(2,), strides=(200,), check_bounds=True)


def test_as_strided_checked_nested_views():
    """Test as_strided with check_bounds=True on a view of a view."""
    a = np.arange(1000, dtype=np.int64)
    b = a[10:100]
    c = b[5:10]

    y = as_strided(c, shape=(2,), strides=(160,), check_bounds=True)
    assert_equal(y.shape, (2,))
    assert_equal(y[0], 15)
    assert_equal(y[1], 35)


def test_as_strided_checked_sliced_array():
    """Test various slicing scenarios."""
    a = np.arange(200, dtype=np.int64)

    b = a[10:20]
    y = as_strided(b, shape=(5,), strides=(16,), check_bounds=True)
    assert_equal(y.shape, (5,))

    c = a[::2]
    y = as_strided(c, shape=(10,), strides=(16,), check_bounds=True)
    assert_equal(y.shape, (10,))


@pytest.mark.parametrize(
    "start,stop,stride_bytes,should_pass",
    [
        (0, 10, 552, True),
        (0, 10, 552 + 1, True),
        (90, 95, 72, True),
        (90, 95, 72 + 1, False),
        (5, 7, -40, True),
        (5, 7, -40 - 1, False),
    ],
)
def test_as_strided_checked_view_parametrized(start, stop, stride_bytes, should_pass):
    """Parametrized test for various view and stride combinations."""
    a = np.arange(100, dtype=np.int64)
    b = a[start:stop]

    if should_pass:
        y = as_strided(b, shape=(2,), strides=(stride_bytes,), check_bounds=True)
        assert_equal(y.shape, (2,))
    else:
        with pytest.raises(ValueError, match="out of bounds"):
            as_strided(b, shape=(2,), strides=(stride_bytes,), check_bounds=True)



"""LU decomposition functions."""

from warnings import warn

from numpy import asarray, asarray_chkfinite
import numpy as np

from scipy._lib._util import _apply_over_batch, _deprecate_dtypes

# Local imports
from ._misc import _datacopied, LinAlgWarning
from .lapack import get_lapack_funcs, _normalize_lapack_dtype, HAS_ILP64
from ._batched_linalg import _lu as _linalg_lu


__all__ = ['lu', 'lu_solve', 'lu_factor']


@_apply_over_batch(('a', 2))
def lu_factor(a, overwrite_a=False, check_finite=True):
    """
    Compute pivoted LU decomposition of a matrix.

    The decomposition is::

        A = P L U

    where P is a permutation matrix, L lower triangular with unit
    diagonal elements, and U upper triangular.

    Parameters
    ----------
    a : (M, N) array_like
        Matrix to decompose
    overwrite_a : bool, optional
        Whether to overwrite data in ``a`` (may improve performance). Default is False.
        See :ref:`tutorial_linalg_overwrite` for details.
    check_finite : bool, optional
        Whether to check that the input matrix contains only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.

    Returns
    -------
    lu : (M, N) ndarray
        Matrix containing U in its upper triangle, and L in its lower triangle.
        The unit diagonal elements of L are not stored.
    piv : (K,) ndarray
        Pivot indices representing the permutation matrix P:
        row i of matrix was interchanged with row piv[i].
        Of shape ``(K,)``, with ``K = min(M, N)``.

    See Also
    --------
    lu : gives lu factorization in more user-friendly format
    lu_solve : solve an equation system using the LU factorization of a matrix

    Notes
    -----
    This is a wrapper to the ``*GETRF`` routines from LAPACK. Unlike
    :func:`lu`, it outputs the L and U factors into a single array
    and returns pivot indices instead of a permutation matrix.

    While the underlying ``*GETRF`` routines return 1-based pivot indices, the
    ``piv`` array returned by ``lu_factor`` contains 0-based indices.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import lu_factor
    >>> A = np.array([[2, 5, 8, 7], [5, 2, 2, 8], [7, 5, 6, 6], [5, 4, 4, 8]])
    >>> lu, piv = lu_factor(A)
    >>> piv
    array([2, 2, 3, 3], dtype=int32)

    Convert LAPACK's ``piv`` array to NumPy index and test the permutation

    >>> def pivot_to_permutation(piv):
    ...     perm = np.arange(len(piv))
    ...     for i in range(len(piv)):
    ...         perm[i], perm[piv[i]] = perm[piv[i]], perm[i]
    ...     return perm
    ...
    >>> p_inv = pivot_to_permutation(piv)
    >>> p_inv
    array([2, 0, 3, 1])
    >>> L, U = np.tril(lu, k=-1) + np.eye(4), np.triu(lu)
    >>> np.allclose(A[p_inv] - L @ U, np.zeros((4, 4)))
    True

    The P matrix in P L U is defined by the inverse permutation and
    can be recovered using argsort:

    >>> p = np.argsort(p_inv)
    >>> p
    array([1, 3, 0, 2])
    >>> np.allclose(A - L[p] @ U, np.zeros((4, 4)))
    True

    or alternatively:

    >>> P = np.eye(4)[p]
    >>> np.allclose(A - P @ L @ U, np.zeros((4, 4)))
    True
    """
    if check_finite:
        a1 = asarray_chkfinite(a)
    else:
        a1 = asarray(a)

    # accommodate empty arrays
    if a1.size == 0:
        lu = np.empty_like(a1)
        piv = np.arange(0, dtype=np.int64 if HAS_ILP64 else np.int32)
        return lu, piv

    overwrite_a = overwrite_a or (_datacopied(a1, a))

    getrf, = get_lapack_funcs(('getrf',), (a1,))
    lu, piv, info = getrf(a1, overwrite_a=overwrite_a)
    if info < 0:
        raise ValueError(
            f'illegal value in {-info}th argument of internal getrf (lu_factor)'
        )
    if info > 0:
        warn(
            f"Diagonal number {info} is exactly zero. Singular matrix.",
            LinAlgWarning,
            stacklevel=2
        )
    return lu, piv


def lu_solve(lu_and_piv, b, trans=0, overwrite_b=False, check_finite=True):
    """Solve an equation system, ``a @ x = b``, given the LU factorization of a.

    The documentation is written assuming array arguments are of specified
    "core" shapes. However, array argument(s) of this function may have additional
    "batch" dimensions prepended to the core shape. In this case, the array is treated
    as a batch of lower-dimensional slices; see :ref:`linalg_batch` for details.

    Parameters
    ----------
    lu_and_piv : tuple
        Factorization of the coefficient matrix a in the form ``(lu, piv)``, as given
        by lu_factor. In particular piv are 0-indexed pivot indices.
    b : array
        Right-hand side
    trans : {0, 1, 2}, optional
        Type of system to solve:

        =====  =========
        trans  system
        =====  =========
        0      a x   = b
        1      a^T x = b
        2      a^H x = b
        =====  =========
    overwrite_b : bool, optional
        Whether to overwrite data in b (may increase performance)
        See :ref:`tutorial_linalg_overwrite` for details.
    check_finite : bool, optional
        Whether to check that the input matrices contain only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.

    Returns
    -------
    x : array
        Solution to the system

    See Also
    --------
    lu_factor : LU factorize a matrix

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import lu_factor, lu_solve
    >>> A = np.array([[2, 5, 8, 7], [5, 2, 2, 8], [7, 5, 6, 6], [5, 4, 4, 8]])
    >>> b = np.array([1, 1, 1, 1])
    >>> lu, piv = lu_factor(A)
    >>> x = lu_solve((lu, piv), b)
    >>> np.allclose(A @ x - b, np.zeros((4,)))
    True

    """
    (lu, piv) = lu_and_piv
    return _lu_solve(lu, piv, b, trans=trans, overwrite_b=overwrite_b,
                     check_finite=check_finite)


@_apply_over_batch(('lu', 2), ('piv', 1), ('b', '1|2'))
def _lu_solve(lu, piv, b, trans, overwrite_b, check_finite):
    if check_finite:
        b1 = asarray_chkfinite(b)
    else:
        b1 = asarray(b)

    _deprecate_dtypes("lu_solve", lu, b)

    overwrite_b = overwrite_b or _datacopied(b1, b)

    if lu.shape[0] != b1.shape[0]:
        raise ValueError(f"Shapes of lu {lu.shape} and b {b1.shape} are incompatible")

    # accommodate empty arrays
    if b1.size == 0:
        m = lu_solve((np.eye(2, dtype=lu.dtype), [0, 1]), np.ones(2, dtype=b.dtype))
        return np.empty_like(b1, dtype=m.dtype)

    getrs, = get_lapack_funcs(('getrs',), (lu, b1))
    x, info = getrs(lu, piv, b1, trans=trans, overwrite_b=overwrite_b)
    if info == 0:
        return x
    raise ValueError(f'illegal value in {-info}th argument of internal gesv|posv')


def lu(a, permute_l=False, overwrite_a=False, check_finite=True,
       p_indices=False):
    """
    Compute LU decomposition of a matrix with partial pivoting.

    The decomposition satisfies::

        A = P @ L @ U

    where ``P`` is a permutation matrix, ``L`` lower triangular with unit
    diagonal elements, and ``U`` upper triangular. If `permute_l` is set to
    ``True`` then ``L`` is returned already permuted and hence satisfying
    ``A = L @ U``.

    Array argument(s) of this function may have additional
    "batch" dimensions prepended to the core shape. In this case, the array is treated
    as a batch of lower-dimensional slices; see :ref:`linalg_batch` for details.

    Parameters
    ----------
    a : (..., M, N) array_like
        Array to decompose
    permute_l : bool, optional
        Perform the multiplication P*L (Default: do not permute)
    overwrite_a : bool, optional
        Whether to overwrite data in a (may improve performance). Default is False.
        See :ref:`tutorial_linalg_overwrite` for details.
    check_finite : bool, optional
        Whether to check that the input matrix contains only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.
    p_indices : bool, optional
        If ``True`` the permutation information is returned as row indices.
        The default is ``False`` for backwards-compatibility reasons.

    Returns
    -------
    (p, l, u) | (pl, u):
        The tuple `(p, l, u)` is returned if `permute_l` is ``False`` (default) else
        the tuple `(pl, u)` is returned, where:

        p : (..., M, M) ndarray
            Permutation arrays or vectors depending on `p_indices`.
        l : (..., M, K) ndarray
            Lower triangular or trapezoidal array with unit diagonal, where the last
            dimension is ``K = min(M, N)``.
        pl : (..., M, K) ndarray
            Permuted L matrix with last dimension being ``K = min(M, N)``.
        u : (..., K, N) ndarray
            Upper triangular or trapezoidal array.

    Notes
    -----
    Permutation matrices are costly since they are nothing but row reorder of
    ``L`` and hence indices are strongly recommended to be used instead if the
    permutation is required. The relation in the 2D case then becomes simply
    ``A = L[P, :] @ U``. In higher dimensions, it is better to use `permute_l`
    to avoid complicated indexing tricks.

    In 2D case, if one has the indices however, for some reason, the
    permutation matrix is still needed then it can be constructed by
    ``np.eye(M)[P, :]``.

    Examples
    --------

    >>> import numpy as np
    >>> from scipy.linalg import lu
    >>> A = np.array([[2, 5, 8, 7], [5, 2, 2, 8], [7, 5, 6, 6], [5, 4, 4, 8]])
    >>> p, l, u = lu(A)
    >>> np.allclose(A, p @ l @ u)
    True
    >>> p  # Permutation matrix
    array([[0., 1., 0., 0.],  # Row index 1
           [0., 0., 0., 1.],  # Row index 3
           [1., 0., 0., 0.],  # Row index 0
           [0., 0., 1., 0.]]) # Row index 2
    >>> p, _, _ = lu(A, p_indices=True)
    >>> p
    array([1, 3, 0, 2], dtype=int32)  # as given by row indices above
    >>> np.allclose(A, l[p, :] @ u)
    True

    We can also use nd-arrays, for example, a demonstration with 4D array:

    >>> rng = np.random.default_rng()
    >>> A = rng.uniform(low=-4, high=4, size=[3, 2, 4, 8])
    >>> p, l, u = lu(A)
    >>> p.shape, l.shape, u.shape
    ((3, 2, 4, 4), (3, 2, 4, 4), (3, 2, 4, 8))
    >>> np.allclose(A, p @ l @ u)
    True
    >>> PL, U = lu(A, permute_l=True)
    >>> np.allclose(A, PL @ U)
    True

    """
    a1 = np.asarray_chkfinite(a) if check_finite else np.asarray(a)
    _deprecate_dtypes("lu", a1)
    if a1.ndim < 2:
        raise ValueError('The input array must be at least two-dimensional.')

    # Also check if dtype is LAPACK compatible
    a1, overwrite_a = _normalize_lapack_dtype(a1, overwrite_a)

    *nd, m, n = a1.shape
    k = min(m, n)
    real_dchar = 'f' if a1.dtype.char in 'fF' else 'd'

    # Empty input
    if min(*a1.shape) == 0:
        if permute_l:
            PL = np.empty(shape=[*nd, m, k], dtype=a1.dtype)
            U = np.empty(shape=[*nd, k, n], dtype=a1.dtype)
            return PL, U
        else:
            P = (np.empty([*nd, 0], dtype=np.int32) if p_indices else
                 np.empty([*nd, 0, 0], dtype=real_dchar))
            L = np.empty(shape=[*nd, m, k], dtype=a1.dtype)
            U = np.empty(shape=[*nd, k, n], dtype=a1.dtype)
            return P, L, U

    # Scalar case
    if a1.shape[-2:] == (1, 1):
        if permute_l:
            return np.ones_like(a1), (a1 if overwrite_a else a1.copy())
        else:
            P = (np.zeros(shape=[*nd, m], dtype=int) if p_indices
                 else np.ones_like(a1))
            return P, np.ones_like(a1), (a1 if overwrite_a else a1.copy())

    P, L, U = _linalg_lu(a1, permute_l, overwrite_a)

    # Convert permutation vecs to permutation arrays
    # permute_l=False needed to enter here to avoid wasted efforts
    if (not p_indices) and (not permute_l):
        if nd:
            Pa = np.zeros([*nd, m, m], dtype=real_dchar)
            # An unreadable index hack - One-hot encoding for perm matrices
            nd_ix = np.ix_(*([np.arange(x) for x in nd]+[np.arange(m)]))
            Pa[(*nd_ix, P)] = 1
            P = Pa
        else:  # 2D case
            Pa = np.zeros([m, m], dtype=real_dchar)
            Pa[np.arange(m), P] = 1
            P = Pa

    return (L, U) if permute_l else (P, L, U)



from _typeshed import Incomplete
from collections.abc import Mapping, Sequence
from types import EllipsisType
from typing import Any, ClassVar, Literal as L, Self, SupportsIndex, overload
from typing_extensions import TypeVar

import numpy as np
from numpy._typing import (
    ArrayLike,
    DTypeLike,
    NDArray,
    _AnyShape,
    _ArrayLikeInt_co,
    _NestedSequence,
    _ShapeLike,
)

__all__ = ["asmatrix", "bmat", "matrix"]

_ShapeT_co = TypeVar("_ShapeT_co", bound=_2D, default=_2D, covariant=True)
_DTypeT_co = TypeVar("_DTypeT_co", bound=np.dtype, default=np.dtype, covariant=True)

type _2D = tuple[int, int]
type _Matrix[ScalarT: np.generic] = matrix[_2D, np.dtype[ScalarT]]
type _ToIndex1 = slice | EllipsisType | NDArray[np.integer | np.bool] | _NestedSequence[int] | None
type _ToIndex2 = tuple[_ToIndex1, _ToIndex1 | SupportsIndex] | tuple[_ToIndex1 | SupportsIndex, _ToIndex1]

###

class matrix(np.ndarray[_ShapeT_co, _DTypeT_co]):
    __array_priority__: ClassVar[float] = 10.0  # pyright: ignore[reportIncompatibleMethodOverride]

    def __new__(cls, data: ArrayLike, dtype: DTypeLike | None = None, copy: bool = True) -> _Matrix[Incomplete]: ...

    #
    @overload  # type: ignore[override]
    def __getitem__(
        self, key: SupportsIndex | _ArrayLikeInt_co | tuple[SupportsIndex | _ArrayLikeInt_co, ...], /
    ) -> Incomplete: ...
    @overload
    def __getitem__(self, key: _ToIndex1 | _ToIndex2, /) -> matrix[_2D, _DTypeT_co]: ...
    @overload
    def __getitem__(self: _Matrix[np.void], key: str, /) -> _Matrix[Incomplete]: ...
    @overload
    def __getitem__(self: _Matrix[np.void], key: list[str], /) -> matrix[_2D, _DTypeT_co]: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    #
    def __mul__(self, other: ArrayLike, /) -> _Matrix[Incomplete]: ...  # type: ignore[override]  # pyright: ignore[reportIncompatibleMethodOverride]
    def __rmul__(self, other: ArrayLike, /) -> _Matrix[Incomplete]: ...  # type: ignore[override]  # pyright: ignore[reportIncompatibleMethodOverride]

    #
    def __pow__(self, other: ArrayLike, /) -> _Matrix[Incomplete]: ...  # type: ignore[override]  # pyright: ignore[reportIncompatibleMethodOverride]
    def __rpow__(self, other: ArrayLike, /) -> _Matrix[Incomplete]: ...  # type: ignore[override]

    # keep in sync with `prod` and `mean`
    @overload  # type: ignore[override]
    def sum(self, axis: None = None, dtype: DTypeLike | None = None, out: None = None) -> Incomplete: ...
    @overload
    def sum(self, axis: _ShapeLike, dtype: DTypeLike | None = None, out: None = None) -> _Matrix[Incomplete]: ...
    @overload
    def sum[OutT: np.ndarray](self, axis: _ShapeLike | None, dtype: DTypeLike | None, out: OutT) -> OutT: ...
    @overload
    def sum[OutT: np.ndarray](self, axis: _ShapeLike | None = None, dtype: DTypeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `sum` and `mean`
    @overload  # type: ignore[override]
    def prod(self, axis: None = None, dtype: DTypeLike | None = None, out: None = None) -> Incomplete: ...
    @overload
    def prod(self, axis: _ShapeLike, dtype: DTypeLike | None = None, out: None = None) -> _Matrix[Incomplete]: ...
    @overload
    def prod[OutT: np.ndarray](self, axis: _ShapeLike | None, dtype: DTypeLike | None, out: OutT) -> OutT: ...
    @overload
    def prod[OutT: np.ndarray](self, axis: _ShapeLike | None = None, dtype: DTypeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `sum` and `prod`
    @overload  # type: ignore[override]
    def mean(self, axis: None = None, dtype: DTypeLike | None = None, out: None = None) -> Incomplete: ...
    @overload
    def mean(self, axis: _ShapeLike, dtype: DTypeLike | None = None, out: None = None) -> _Matrix[Incomplete]: ...
    @overload
    def mean[OutT: np.ndarray](self, axis: _ShapeLike | None, dtype: DTypeLike | None, out: OutT) -> OutT: ...
    @overload
    def mean[OutT: np.ndarray](self, axis: _ShapeLike | None = None, dtype: DTypeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `var`
    @overload  # type: ignore[override]
    def std(self, axis: None = None, dtype: DTypeLike | None = None, out: None = None, ddof: float = 0) -> Incomplete: ...
    @overload
    def std(self, axis: _ShapeLike, dtype: DTypeLike | None = None, out: None = None, ddof: float = 0) -> _Matrix[Incomplete]: ...
    @overload
    def std[OutT: np.ndarray](self, axis: _ShapeLike | None, dtype: DTypeLike | None, out: OutT, ddof: float = 0) -> OutT: ...
    @overload
    def std[OutT: np.ndarray](  # pyright: ignore[reportIncompatibleMethodOverride]
        self, axis: _ShapeLike | None = None, dtype: DTypeLike | None = None, *, out: OutT, ddof: float = 0
    ) -> OutT: ...

    # keep in sync with `std`
    @overload  # type: ignore[override]
    def var(self, axis: None = None, dtype: DTypeLike | None = None, out: None = None, ddof: float = 0) -> Incomplete: ...
    @overload
    def var(self, axis: _ShapeLike, dtype: DTypeLike | None = None, out: None = None, ddof: float = 0) -> _Matrix[Incomplete]: ...
    @overload
    def var[OutT: np.ndarray](self, axis: _ShapeLike | None, dtype: DTypeLike | None, out: OutT, ddof: float = 0) -> OutT: ...
    @overload
    def var[OutT: np.ndarray](  # pyright: ignore[reportIncompatibleMethodOverride]
        self, axis: _ShapeLike | None = None, dtype: DTypeLike | None = None, *, out: OutT, ddof: float = 0
    ) -> OutT: ...

    # keep in sync with `all`
    @overload  # type: ignore[override]
    def any(self, axis: None = None, out: None = None) -> np.bool: ...
    @overload
    def any(self, axis: _ShapeLike, out: None = None) -> _Matrix[np.bool]: ...
    @overload
    def any[OutT: np.ndarray](self, axis: _ShapeLike | None, out: OutT) -> OutT: ...
    @overload
    def any[OutT: np.ndarray](self, axis: _ShapeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `any`
    @overload  # type: ignore[override]
    def all(self, axis: None = None, out: None = None) -> np.bool: ...
    @overload
    def all(self, axis: _ShapeLike, out: None = None) -> _Matrix[np.bool]: ...
    @overload
    def all[OutT: np.ndarray](self, axis: _ShapeLike | None, out: OutT) -> OutT: ...
    @overload
    def all[OutT: np.ndarray](self, axis: _ShapeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `min` and `ptp`
    @overload  # type: ignore[override]
    def max[ScalarT: np.generic](self: NDArray[ScalarT], axis: None = None, out: None = None) -> ScalarT: ...
    @overload
    def max(self, axis: _ShapeLike, out: None = None) -> matrix[_2D, _DTypeT_co]: ...
    @overload
    def max[OutT: np.ndarray](self, axis: _ShapeLike | None, out: OutT) -> OutT: ...
    @overload
    def max[OutT: np.ndarray](self, axis: _ShapeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `max` and `ptp`
    @overload  # type: ignore[override]
    def min[ScalarT: np.generic](self: NDArray[ScalarT], axis: None = None, out: None = None) -> ScalarT: ...
    @overload
    def min(self, axis: _ShapeLike, out: None = None) -> matrix[_2D, _DTypeT_co]: ...
    @overload
    def min[OutT: np.ndarray](self, axis: _ShapeLike | None, out: OutT) -> OutT: ...
    @overload
    def min[OutT: np.ndarray](self, axis: _ShapeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `max` and `min`
    @overload
    def ptp[ScalarT: np.generic](self: NDArray[ScalarT], axis: None = None, out: None = None) -> ScalarT: ...
    @overload
    def ptp(self, axis: _ShapeLike, out: None = None) -> matrix[_2D, _DTypeT_co]: ...
    @overload
    def ptp[OutT: np.ndarray](self, axis: _ShapeLike | None, out: OutT) -> OutT: ...
    @overload
    def ptp[OutT: np.ndarray](self, axis: _ShapeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `argmin`
    @overload  # type: ignore[override]
    def argmax[ScalarT: np.generic](self: NDArray[ScalarT], axis: None = None, out: None = None) -> np.intp: ...
    @overload
    def argmax(self, axis: _ShapeLike, out: None = None) -> _Matrix[np.intp]: ...
    @overload
    def argmax[OutT: NDArray[np.integer | np.bool]](self, axis: _ShapeLike | None, out: OutT) -> OutT: ...
    @overload
    def argmax[OutT: NDArray[np.integer | np.bool]](self, axis: _ShapeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # keep in sync with `argmax`
    @overload  # type: ignore[override]
    def argmin[ScalarT: np.generic](self: NDArray[ScalarT], axis: None = None, out: None = None) -> np.intp: ...
    @overload
    def argmin(self, axis: _ShapeLike, out: None = None) -> _Matrix[np.intp]: ...
    @overload
    def argmin[OutT: NDArray[np.integer | np.bool]](self, axis: _ShapeLike | None, out: OutT) -> OutT: ...
    @overload
    def argmin[OutT: NDArray[np.integer | np.bool]](self, axis: _ShapeLike | None = None, *, out: OutT) -> OutT: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # the second overload handles the (rare) case that the matrix is not 2-d
    @overload
    def tolist[T](self: _Matrix[np.generic[T]]) -> list[list[T]]: ...  # pyright: ignore[reportIncompatibleMethodOverride]
    @overload
    def tolist(self) -> Incomplete: ...  # pyright: ignore[reportIncompatibleMethodOverride]

    # these three methods will at least return a `2-d` array of shape (1, n)
    def squeeze(self, /, axis: _ShapeLike | None = None) -> matrix[_2D, _DTypeT_co]: ...
    def ravel(self, /, order: L["K", "A", "C", "F"] | None = "C") -> matrix[_2D, _DTypeT_co]: ...  # type: ignore[override]
    def flatten(self, /, order: L["K", "A", "C", "F"] | None = "C") -> matrix[_2D, _DTypeT_co]: ...  # type: ignore[override]

    # matrix.T is inherited from _ScalarOrArrayCommon
    def getT(self) -> Self: ...
    @property
    def I(self) -> _Matrix[Incomplete]: ...  # noqa: E743
    def getI(self) -> _Matrix[Incomplete]: ...
    @property
    def A(self) -> np.ndarray[_2D, _DTypeT_co]: ...
    def getA(self) -> np.ndarray[_2D, _DTypeT_co]: ...
    @property
    def A1(self) -> np.ndarray[_AnyShape, _DTypeT_co]: ...
    def getA1(self) -> np.ndarray[_AnyShape, _DTypeT_co]: ...
    @property
    def H(self) -> matrix[_2D, _DTypeT_co]: ...
    def getH(self) -> matrix[_2D, _DTypeT_co]: ...

def bmat(
    obj: str | Sequence[ArrayLike] | NDArray[Any],
    ldict: Mapping[str, Any] | None = None,
    gdict: Mapping[str, Any] | None = None,
) -> _Matrix[Incomplete]: ...

def asmatrix(data: ArrayLike, dtype: DTypeLike | None = None) -> _Matrix[Incomplete]: ...



import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest
import matplotlib.pyplot as plt
from matplotlib.testing.decorators import image_comparison
import matplotlib.transforms as mtransforms


def velocity_field():
    Y, X = np.mgrid[-3:3:100j, -3:3:200j]
    U = -1 - X**2 + Y
    V = 1 + X - Y**2
    return X, Y, U, V


def swirl_velocity_field():
    x = np.linspace(-3., 3., 200)
    y = np.linspace(-3., 3., 100)
    X, Y = np.meshgrid(x, y)
    a = 0.1
    U = np.cos(a) * (-Y) - np.sin(a) * X
    V = np.sin(a) * (-Y) + np.cos(a) * X
    return x, y, U, V


@image_comparison(['streamplot_startpoints.png'], remove_text=True, style='mpl20',
                  tol=0.003)
def test_startpoints():
    # Test varying startpoints. Also tests a non-default num_arrows argument.
    X, Y, U, V = velocity_field()
    start_x, start_y = np.meshgrid(np.linspace(X.min(), X.max(), 5),
                                   np.linspace(Y.min(), Y.max(), 5))
    start_points = np.column_stack([start_x.ravel(), start_y.ravel()])
    plt.streamplot(X, Y, U, V, start_points=start_points, num_arrows=4)
    plt.plot(start_x, start_y, 'ok')


@image_comparison(['streamplot_colormap.png'], remove_text=True, style='mpl20',
                  tol=0.022)
def test_colormap():
    X, Y, U, V = velocity_field()
    plt.streamplot(X, Y, U, V, color=U, density=0.6, linewidth=2,
                   cmap="autumn")
    plt.colorbar()


@image_comparison(['streamplot_linewidth.png'], remove_text=True, style='mpl20',
                  tol=0.03)
def test_linewidth():
    X, Y, U, V = velocity_field()
    speed = np.hypot(U, V)
    lw = 5 * speed / speed.max()
    ax = plt.figure().subplots()
    ax.streamplot(X, Y, U, V, density=[0.5, 1], color='k', linewidth=lw, num_arrows=2)


@image_comparison(['streamplot_masks_and_nans.png'],
                  remove_text=True, style='mpl20')
def test_masks_and_nans():
    X, Y, U, V = velocity_field()
    mask = np.zeros(U.shape, dtype=bool)
    mask[40:60, 80:120] = 1
    U[:20, :40] = np.nan
    U = np.ma.array(U, mask=mask)
    ax = plt.figure().subplots()
    with np.errstate(invalid='ignore'):
        ax.streamplot(X, Y, U, V, color=U, cmap="Blues")


@image_comparison(['streamplot_maxlength.png'],
                  remove_text=True, style='mpl20', tol=0.302)
def test_maxlength():
    x, y, U, V = swirl_velocity_field()
    ax = plt.figure().subplots()
    ax.streamplot(x, y, U, V, maxlength=10., start_points=[[0., 1.5]],
                  linewidth=2, density=2)
    assert ax.get_xlim()[-1] == ax.get_ylim()[-1] == 3
    # Compatibility for old test image
    ax.set(xlim=(None, 3.2555988021882305), ylim=(None, 3.078326760195413))


@image_comparison(['streamplot_maxlength_no_broken.png'],
                  remove_text=True, style='mpl20', tol=0.302)
def test_maxlength_no_broken():
    x, y, U, V = swirl_velocity_field()
    ax = plt.figure().subplots()
    ax.streamplot(x, y, U, V, maxlength=10., start_points=[[0., 1.5]],
                  linewidth=2, density=2, broken_streamlines=False)
    assert ax.get_xlim()[-1] == ax.get_ylim()[-1] == 3
    # Compatibility for old test image
    ax.set(xlim=(None, 3.2555988021882305), ylim=(None, 3.078326760195413))


@image_comparison(['streamplot_direction.png'],
                  remove_text=True, style='mpl20', tol=0.073)
def test_direction():
    x, y, U, V = swirl_velocity_field()
    plt.streamplot(x, y, U, V, integration_direction='backward',
                   maxlength=1.5, start_points=[[1.5, 0.]],
                   linewidth=2, density=2)


@image_comparison(['streamplot_integration.png'], style='mpl20', tol=0.05)
def test_integration_options():
    # Linear potential flow over a lifting cylinder
    n = 50
    x, y = np.meshgrid(np.linspace(-2, 2, n), np.linspace(-3, 3, n))
    th = np.arctan2(y, x)
    r = np.sqrt(x**2 + y**2)
    vr = -np.cos(th) / r**2
    vt = -np.sin(th) / r**2 - 1 / r
    vx = vr * np.cos(th) - vt * np.sin(th) + 1.0
    vy = vr * np.sin(th) + vt * np.cos(th)

    # Seed points
    n_seed = 50
    seed_pts = np.column_stack((np.full(n_seed, -1.75), np.linspace(-2, 2, n_seed)))

    fig, axs = plt.subplots(3, 1, figsize=(6, 14))
    th_circ = np.linspace(0, 2 * np.pi, 100)
    for ax, max_val in zip(axs, [0.05, 1, 5]):
        ax_ins = ax.inset_axes([0.0, 0.7, 0.3, 0.35])
        for ax_curr, is_inset in zip([ax, ax_ins], [False, True]):
            ax_curr.streamplot(
                x,
                y,
                vx,
                vy,
                start_points=seed_pts,
                broken_streamlines=False,
                arrowsize=1e-10,
                linewidth=2 if is_inset else 0.6,
                color="k",
                integration_max_step_scale=max_val,
                integration_max_error_scale=max_val,
            )

            # Draw the cylinder
            ax_curr.fill(
                np.cos(th_circ),
                np.sin(th_circ),
                color="w",
                ec="k",
                lw=6 if is_inset else 2,
            )

            # Set axis properties
            ax_curr.set_aspect("equal")

        # Set axis limits and show zoomed region
        ax_ins.set_xlim(-1.2, -0.7)
        ax_ins.set_ylim(-0.8, -0.4)
        ax_ins.set_yticks(())
        ax_ins.set_xticks(())

        ax.set_ylim(-1.5, 1.5)
        ax.axis("off")
        ax.indicate_inset_zoom(ax_ins, ec="k")

    fig.tight_layout()


def test_streamplot_limits():
    ax = plt.axes()
    x = np.linspace(-5, 10, 20)
    y = np.linspace(-2, 4, 10)
    y, x = np.meshgrid(y, x)
    trans = mtransforms.Affine2D().translate(25, 32) + ax.transData
    plt.barbs(x, y, np.sin(x), np.cos(y), transform=trans)
    # The calculated bounds are approximately the bounds of the original data,
    # this is because the entire path is taken into account when updating the
    # datalim.
    assert_array_almost_equal(ax.dataLim.bounds, (20, 30, 15, 6),
                              decimal=1)


def test_streamplot_grid():
    u = np.ones((2, 2))
    v = np.zeros((2, 2))

    # Test for same rows and columns
    x = np.array([[10, 20], [10, 30]])
    y = np.array([[10, 10], [20, 20]])

    with pytest.raises(ValueError, match="The rows of 'x' must be equal"):
        plt.streamplot(x, y, u, v)

    x = np.array([[10, 20], [10, 20]])
    y = np.array([[10, 10], [20, 30]])

    with pytest.raises(ValueError, match="The columns of 'y' must be equal"):
        plt.streamplot(x, y, u, v)

    x = np.array([[10, 20], [10, 20]])
    y = np.array([[10, 10], [20, 20]])
    plt.streamplot(x, y, u, v)

    # Test for maximum dimensions
    x = np.array([0, 10])
    y = np.array([[[0, 10]]])

    with pytest.raises(ValueError, match="'y' can have at maximum "
                                         "2 dimensions"):
        plt.streamplot(x, y, u, v)

    # Test for equal spacing
    u = np.ones((3, 3))
    v = np.zeros((3, 3))
    x = np.array([0, 10, 20])
    y = np.array([0, 10, 30])

    with pytest.raises(ValueError, match="'y' values must be equally spaced"):
        plt.streamplot(x, y, u, v)

    # Test for strictly increasing
    x = np.array([0, 20, 40])
    y = np.array([0, 20, 10])


def test_streamplot_integration_params():
    x = np.array([[10, 20], [10, 20]])
    y = np.array([[10, 10], [20, 20]])
    u = np.ones((2, 2))
    v = np.zeros((2, 2))

    err_str = "The value of integration_max_step_scale must be > 0, got -0.5"
    with pytest.raises(ValueError, match=err_str):
        plt.streamplot(x, y, u, v, integration_max_step_scale=-0.5)

    err_str = "The value of integration_max_error_scale must be > 0, got 0.0"
    with pytest.raises(ValueError, match=err_str):
        plt.streamplot(x, y, u, v, integration_max_error_scale=0.0)


def test_streamplot_inputs():  # test no exception occurs.
    # fully-masked
    plt.streamplot(np.arange(3), np.arange(3),
                   np.full((3, 3), np.nan), np.full((3, 3), np.nan),
                   color=np.random.rand(3, 3))
    # array-likes
    plt.streamplot(range(3), range(3),
                   np.random.rand(3, 3), np.random.rand(3, 3))
